#!/bin/bash

# Copyright (c) 2020 by Pulse Secure, LLC. All rights reserved

HOMEDIR=$HOME
INSTALLDIR=/usr/local/pulse
PULSEDIR=$HOME/.pulsesecure/pulse
PULSECERTDIR=$PULSEDIR/certificates
PRIVATESUBDIR=.private
args=""
ive_ip=""
NOARGS=$#

unset -f eval
unset -f echo
unset -f exit
unset -f read
unset -f printf
unset -f shift

if [ -e /usr/bin/echo ]; then
  ECHO=/usr/bin/echo
elif [ -e /bin/echo ]; then
  ECHO=/bin/echo
elif [ -e /usr/sbin/echo ]; then
  ECHO=/usr/sbin/echo
else
  printf "Unable to find absolute path for echo command. Exiting" 
  exit 5
fi

full_path_command() {
    cmd="$1"
    # Check if command exists in /usr/bin
    if [ -e "/usr/bin/$cmd" ]; then
        $ECHO "/usr/bin/$cmd"
        return
    fi
    # Check if command exists in /usr/sbin
    if [ -e "/usr/sbin/$cmd" ]; then
        $ECHO "/usr/sbin/$cmd"
        return
    fi
    # Check if command exists in /bin
    if [ -e "/bin/$cmd" ]; then
        $ECHO "/bin/$cmd"
        return
    fi
	$ECHO "$cmd command not found in /usr/bin, /usr/sbin or /bin."
    $ECHO ""
}
BASENAME=$(full_path_command "basename")
READLINK=$(full_path_command "readlink")
CP=$(full_path_command "cp")
MKDIR=$(full_path_command "mkdir")
RM=$(full_path_command "rm")
LS=$(full_path_command "ls")
MV=$(full_path_command "mv")
CHMOD=$(full_path_command "chmod")
CHOWN=$(full_path_command "chown")
DF=$(full_path_command "df")
DATE=$(full_path_command "date")
GREP=$(full_path_command "grep")
AWK=$(full_path_command "awk")
SED=$(full_path_command "sed")
CERTUTIL=$(full_path_command "certutil")
PK12UTIL=$(full_path_command "pk12util")
OPENSSLCMD=$(full_path_command "openssl")
SCRNAME=$($BASENAME $0)

# The below code takes all the script invoke argsand remove meta-characters $#``\ and all other escape characters
#It allows only A-Z,a-z,0-9,dot,hypen,forward slash and underscore
#replaces double dots and forward slash with single dots or slash
sanitized_input=()
for arg in "$@"; do
    sanitized_arg=$(printf '%s' "$arg" | "$SED"  's/[^a-zA-Z0-9_/.-]//g')
    sanitized_arg=$(printf '%s' "$sanitized_arg" | "$SED" 's/\.\{2,\}/\./g')
	sanitized_arg=$(printf '%s' "$sanitized_arg" | "$SED" 's#/\{2,\}#/#g')
    sanitized_input+=("$sanitized_arg")
done

$ECHO "Sanitized Arguments" "${sanitized_input[@]}"
params_len="${#sanitized_input[@]}"
$ECHO "Sanitized Arguments Length $params_len"


function check_error()
{
	errorCode=$1
	errorString=$2
	if [ $1 != 0 ] && [ "X$errorString" != "X" ]; then 
		$ECHO "ErrorMessage : $errorString"
		exit 3
	fi
}

function install_pfx()
{
	filename=$1
	pfxFilepath=$($READLINK -f "$filename")
	keyFileBaseName=$($BASENAME "$filename")
	keyFileName="${keyFileBaseName%.*}"
	privKeyFileName="$PULSECERTDIR/$PRIVATESUBDIR/$keyFileName.key"
	pubKeyFileName="$PULSECERTDIR/$keyFileName.pem"
	pubKeyTmpFileName="$PULSECERTDIR/$keyFileName-tmp-pub.pem"

	# pkcs12 file format support starts here
	if [ ! -f "$filename" ]; then 
		$ECHO "$filename does not exists. Please check the pfx file location "
		exit 2;
	fi

    warn_user_for_overwrite "${pubKeyFileName}" 
	#$OPENSSLCMD pkcs12 -info -in $filename -passin pass:$password -nodes 2>/dev/null
	#$OPENSSLCMD pkcs12 -info -in $filename -nodes 2>/dev/null
	#check_error $? "$FUNCNAME: File Extension is .pfx/.p12 but content is not"
	$ECHO "Certificates will be installed for Pulse Certificate store and CEF Certificate store."
    $ECHO -n Enter Import Password:
    	read -s paswd

	$ECHO "Extracting Public Key from $filename"
	pubKeyExtractCmd=($OPENSSLCMD pkcs12 -in $filename -clcerts -nokeys -out $pubKeyTmpFileName -nodes -passin pass:"$paswd" -passout pass:)
	"${pubKeyExtractCmd[@]}"
	ret=$?
	if [ $ret != 0 ]; then
		if [ -e $pubKeyTmpFileName ]
		then
			$RM "$pubKeyTmpFileName"
		fi
	fi
	check_error $ret "$FUNCNAME: Public key extraction failed"

	check_already_installed "${pubKeyTmpFileName}"
	$MV $pubKeyTmpFileName $pubKeyFileName
	$ECHO "Extracting Private Key from $filename"
	privKeyExtractCmd=($OPENSSLCMD pkcs12 -in "$filename" -nocerts -out "$privKeyFileName" -nodes  -passin pass:"$paswd" -passout pass:)
	"${privKeyExtractCmd[@]}"
	ret=$?
	if [ $ret != 0 ]; then
		$RM "$privKeyFileName"
		$RM "$pubKeyFileName"
	fi
	check_error $ret "$FUNCNAME: Private key extraction failed"
    $ECHO "Successfully added certificate to Pulse Certificate store."
    
    pk12CMD=(${PK12UTIL} -d sql:$HOME/.pki/nssdb  -i "${filename}"  -W "${paswd}")
	"${pk12CMD[@]}"
    ret=$?
    if [ $ret != 0 ]; then
        $ECHO "Failed to add certificate to CEF Certificate store."
    else
        $ECHO "Successfully added certificate to CEF Certificate store."
	fi
}

function warn_user_for_overwrite()
{
    certFile=$1
    if [ -f "$certFile" ]; then
        name=$($BASENAME "$certFile" ".pem")
        $ECHO
		$ECHO "Client certificate with name ${name} already"\
             "exists in pulse certificate store."
        read -e -p "Do you want to continue[y/n]: " choice
        if ! [[ "${choice:0:1}" == "Y" || "${choice:0:1}" == "y" ]]; then
            $ECHO "Aborting the certificate installation."
            exit 0;
        fi
    fi
}

function check_already_installed()
{
	certFile=$1
	if $LS $PULSECERTDIR/*.pem &>/dev/null
	then
		opensslCNCmd='$OPENSSLCMD x509 -noout -subject_hash -in '\"$certFile\"''
		CNHashOld=$(eval $opensslCNCmd 2>&1 )
   		for i in $PULSECERTDIR/*.pem;
		do
			# skip checking against the same file 
			if [ $certFile != $i ]; then
			openHashNew='$OPENSSLCMD x509 -noout -subject_hash -in '\"$i\"''
				CNHashNew=$(eval $openHashNew 2>&1 )
				if [ "$CNHashOld" = "$CNHashNew" ]; then
					$ECHO "Certificate is already present in pulse certificate store. Aborting the certificate installation."
					$RM $certFile
					exit 0;
				fi
			fi
		done
	fi
}

function check_cert_names_same()
{
    priv="$1"
    pub="$2"
    priv=$($BASENAME "$priv")
    priv=${priv%.*}
    pub=$($BASENAME "$pub")
    pub=${pub%.*}
    if [ "$priv" != "$pub" ]; then
        $ECHO "Failed to install certificate. Both Private($priv) and Public($pub) certificate should have same name."
        exit 0;
    fi
}

function install_keys()
{
	FILETYPE=$1
	privKeyInFile="$2"
	privKeyOutFile="$3"
	FAIL=1
	keytype="rsa dsa"
	for i in `$ECHO $keytype`
	do
		installKeyCmd='$OPENSSLCMD '$i' -inform '$FILETYPE' -in '\"$privKeyInFile\"' -out '\"$privKeyOutFile\"' 2>/dev/null'
		eval $installKeyCmd
		if [ $? == 0 ]; then
			FAIL=0
			break;
		fi
	done
	check_error $FAIL "Failed to extract private keys. Supported keys are rsa and dsa"
}

function install_priv_pub_keys()
{
	$ECHO "$1"
	$ECHO "$2"
	privKeyFilePath=$1
	privKeyFileBaseName=$($BASENAME "$1")
	privKeyFileName="${privKeyFileBaseName%.*}"
	privKeyFileExt="${privKeyFileBaseName##*.}"
	pubKeyFilePath=$2
	pubKeyFileBaseName=$($BASENAME "$2")
	pubKeyFileName="${pubKeyFileBaseName%.*}"
	pubKeyFileExt="${pubKeyFileBaseName##*.}"
	privKeyPEMFile="$PULSECERTDIR/$PRIVATESUBDIR/${privKeyFileName}.key"
	pubKeyPEMFile="$PULSECERTDIR/$pubKeyFileName.pem"
	pfxFile="$PULSECERTDIR/$pubKeyFileName.pfx"

	filepath=$($READLINK -f "$pubKeyFilePath")
	check_already_installed "$filepath"
    warn_user_for_overwrite "${pubKeyPEMFile}"
	# Public Key Handling 
	if [[ $pubKeyFileExt == *"p7b" || $pubKeyFileExt == *"p7c" ]]; then
		pkcs7Cmd='$OPENSSLCMD pkcs7 -print_certs -in '\"$pubKeyFilePath\"' -out '\"$pubKeyPEMFile\"
		eval $pkcs7Cmd
		check_error $? "$FUNCNAME: convert $pubKeyFileName to PEM format failed"
	else
		# pkcs 8 format should be given as pem/der file here
		if [[ $pubKeyFileExt == *"der" || $pubKeyFileExt == *"cer" ]]; then
			x509Cmd='$OPENSSLCMD x509 -inform der -in '\"$pubKeyFilePath\"' -out '\"$pubKeyPEMFile\"
			eval $x509Cmd
			check_error $? "$FUNCNAME: convert $pubKeyFileName to PEM format failed"
		elif [[ $pubKeyFileExt == *"pem" || $pubKeyFileExt == *"crt" ||
				$pubKeyFileExt == *"key" || $pubKeyFileExt == *"pub" ]]; then
			$CP "$pubKeyFilePath" "$pubKeyPEMFile"
		else
			check_error 1 "$FUNCNAME: Unknown Public Key File Format"
		fi
	fi
	# Private Key Handling 
	if [[ $privKeyFileExt == *"der" || $privKeyFileExt == *"cer" ]]; then
		install_keys "der" "$privKeyFilePath" "$privKeyPEMFile"
	elif [[ $privKeyFileExt == *"pem" || $privKeyFileExt == *"crt" ||
				$privKeyFileExt == *"key" ]]; then
		# this command removes the password temporarily to install it in gnome-keyring
		install_keys "pem" "$privKeyFilePath" "$privKeyPEMFile"
	elif [[ $privKeyFileExt == *"pk8" ]]; then
		install_keys "pkcs8" "$privKeyFilePath" "$privKeyPEMFile"
	else
		check_error 1 "$FUNCNAME: Unknown Private Key File Format"
	fi
    $ECHO "Successfully added certificate to Pulse Certificate store."
    cmd=($OPENSSLCMD pkcs12 -export -in "$pubKeyPEMFile" -inkey "$privKeyPEMFile" -out "$pfxFile" -passin pass: -passout pass:)
    "${cmd[@]}"
    ret=$?
    if [ $ret == 0 ]; then
	paswd=
    	pk12CMD=(${PK12UTIL} -d sql:$HOME/.pki/nssdb  -i "${pfxFile}"  -W "${paswd}")
		"${pk12CMD[@]}"
    	ret=$?
    	if [ $ret != 0 ]; then
        	$ECHO  "Failed to add certificate to CEF certificate store."
		else
            $ECHO  "Successfully added certificate to CEF Certificate store."
		fi	
    else
		$ECHO  "Failed to convert certificate $2 to pfx format."
	fi

}
function keyUsage() 
{
    $ECHO  "Run command line client Options:"
        $ECHO  "    Client Certificate Options:"
        $ECHO  "        $SCRNAME install_certificates "
        $ECHO  "            [-inpfx < PFX file > ]"
        $ECHO  "            [-inpriv <private file> -inpub <public file>]"
		$ECHO  "        Note: password is required for installing private and public keys separately."
        $ECHO  "                                                                                         "
        $ECHO  "        $SCRNAME delete_certificates "
        $ECHO  "            [-certName <Certificate Name>]"
        $ECHO  "        Note: To delete certificates installed in CEF certificate store, run the following command."
        $ECHO  "              where, 'Nickname' can be found from '$SCRNAME list_installed_certificates'"
        $ECHO  "        /usr/bin/certutil -d sql:${HOME}/.pki/nssdb -D -n <Nickname>"
        $ECHO  "        $SCRNAME list_installed_certificates "
        exit 1
}
######################################################################################################
# Function to install certificates
# Args   : certificate details
# Return : None
# function install_certificate ()
function install_certificate()
{
        $ECHO  "Certficate is installing by user: \"$USER\" "\
             "Please make sure that client certificates to be installed by logged in user only."
        read -e -p "Do you want to continue[y/n]: " choice

        if ! [[ "${choice:0:1}" == "Y" || "${choice:0:1}" == "y" ]]; then
            $ECHO  "Aborting the certificate installation."
            exit 0;
        fi

	privKeyFileName=""
	pubKeyFileName=""
	filename=""
	filename="$1"
	privKeyFileName="$2"
	pubKeyFileName="$3"

	$ECHO "filename $filename"
	$ECHO "privKeyFileName $privKeyFileName"
	$ECHO "pubKeyFileName $pubKeyFileName"

	if [ ! -d $PULSECERTDIR ]; then
		$ECHO  "$PULSECERTDIR does not exists. Creating.."
		$MKDIR -p $PULSECERTDIR
	fi

	if [ ! -d $PULSECERTDIR/$PRIVATESUBDIR ]; then
		$ECHO  "$PULSECERTDIR does not exists. Creating.."
		$MKDIR -p $PULSECERTDIR/$PRIVATESUBDIR
	fi

    if [ ! -d ${HOME}/.pki/nssdb ]; then
        $MKDIR -p ${HOME}/.pki/nssdb
        ${CERTUTIL} -N -d ${HOME}/.pki/nssdb --empty-password
    fi
   

	if [[ $filename == *".pfx" || $filename == *".p12" ]]
	then
		install_pfx "$filename"
	elif [ "X$privKeyFileName" != "X" ] && [ "X$pubKeyFileName" != "X" ]; then 
		$ECHO  "Private Key: $privKeyFileName and Public Key: $pubKeyFileName"
        check_cert_names_same "$privKeyFileName" "$pubKeyFileName"
		install_priv_pub_keys "$privKeyFileName" "$pubKeyFileName"
	else 
		keyUsage
	fi
	#make certificate install directory only for the logged in user.
	$CHMOD -R go-rwx $PULSECERTDIR
	$CHMOD -R go-rwx ${HOME}/.pki
}
# End of function install_certificate ()
######################################################################################################
######################################################################################################
# Function to delete certificates
# Args   : certificate name 
# Return : None
# function delete_certificate ()
function delete_certificate()
{
	cert_name="$1"
	if [ "X$cert_name" != "X" ]; then 
		$ECHO  "Certificate Name :$cert_name "
		if [ -e $PULSECERTDIR/"$cert_name".pem ]; then 
			$RM -rf $PULSECERTDIR/"$cert_name".pem
		else
			$ECHO  -e "Public key file $PULSECERTDIR/$cert_name.pem doesn't exists"
		fi
	else
		keyUsage
	fi
}
# End of function delete_certificate ()

#List the installed certificate in pulse certficate store.
function list_certificates()
{
	if $LS $PULSECERTDIR/*.pem &>/dev/null
	
	then
   		for i in $PULSECERTDIR/*.pem;
		do
		    name=$($BASENAME "$i" ".pem")
			$ECHO  -e "\nCertificate Name:" $name;
			opensslListCmd='$OPENSSLCMD x509 -in '\"$i\"' -text | $GREP -i "Subject:\|Issuer:\|Validity\|Not Before\|Not After";'
  			eval $opensslListCmd
		done
	else
	$ECHO  "No Certificates found."
	fi
    if [ -d ${HOME}/.pki ]; then
    $ECHO  "###############################################################################"
	$ECHO  "Certificates installed for CEF Certificate Store:"
    	${CERTUTIL} -d sql:$HOME/.pki/nssdb -L
    fi
}

######################################################################################################


if [ "X$1" = "Xinstall_certificates" ] ; then
inpfx_value=""
inpriv_value=""
inpub_value=""

# Initialize counter
i=1
while [ "$i" -lt "$params_len" ]; do
    key="${sanitized_input[i]}"
    case $key in
        -inpfx)
            inpfx_value="${sanitized_input[i+1]}"
               ((i=i+2))  ;;
        -inpriv)
            inpriv_value="${sanitized_input[i+1]}"
                 ((i=i+2)) ;;
        -inpub)
            inpub_value="${sanitized_input[i+1]}"
                ((i=i+2)) ;;
        *)
		keyUsage
        break
    esac
done
    install_certificate "$inpfx_value" "$inpriv_value" "$inpub_value"
elif [ "X$1" = "Xdelete_certificates" ] ; then
if [ "$#" -ne 3 ] || [ "X$2" != "X-certName" ]; then
        $ECHO "Usage certificate_installer.sh delete_certificates -certName <Certificate Name>"
		exit 8
else
    delete_certificate "${sanitized_input[2]}"
fi   
elif [ "X$1" = "Xlist_installed_certificates" ] ; then
if [ $# -gt 1 ]; then
    $ECHO "Listing Installed Certificates does not require any parmeter. Please enter exact command."
    exit 1
fi
    list_certificates
else
    keyUsage
fi
